import copy
import torch as th
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import gymnasium as gym
from runner import get_offpolicy_runner
from rainbow_naf.noisy_linear import NoisyLinear

"""
Rainbow components:
1. Dueling: done as part of naf
2. Double DQN: done, greatly improves performance
3. Noisy dqn: done, but needs to be tuned and tried on more complex environments
4. Prioritised replay: should be easy to re-implement
5. Distributional: might be difficult given the nature of the Q-network's architecture
6. N-step learning:
"""

# env = gym.make('gym_cartpole_continuous:CartPoleContinuous-v1')
env = gym.make('LunarLanderContinuous-v2')

runner = get_offpolicy_runner(
	env,
	'naf', 
	freq=1,
	capacity=200000, 
	batch_size=128, 
	train_after=1000)

class NAF(nn.Module):
	"""
	source:
	https://github.com/BY571/Normalized-Advantage-Function-NAF-
	"""
	def __init__(self, state_size, action_size, layer_size):
		super(NAF, self).__init__()
		self.input_shape = state_size
		self.action_size = action_size
				
		self.head_1 = NoisyLinear(self.input_shape, layer_size)
		# self.bn1 = nn.BatchNorm1d(layer_size)
		self.ff_1 = NoisyLinear(layer_size, layer_size)
		# self.bn2 = nn.BatchNorm1d(layer_size)
		self.action_values = nn.Linear(layer_size, action_size)
		self.value = nn.Linear(layer_size, 1)
		self.matrix_entries = nn.Linear(layer_size, int(self.action_size*(self.action_size+1)/2))     
	
	def reset_noise(self):
		"""Reset all noisy layers."""
		self.head_1.reset_noise()
		self.ff_1.reset_noise()
		# self.action_values.reset_noise()
		# self.value.reset_noise()
		# self.matrix_entries.reset_noise() 
	
	def forward(self, input_, action=None):
		x = th.relu(self.head_1(input_))
		# x = self.bn1(x)
		x = th.relu(self.ff_1(x))
		# x = self.bn2(x)
		action_value = th.tanh(self.action_values(x))
		entries = th.tanh(self.matrix_entries(x))
		V = self.value(x)
		
		action_value = action_value.unsqueeze(-1)
		
		# create lower-triangular matrix
		L = th.zeros((input_.shape[0], self.action_size, self.action_size)).to(input_.device)

		# get lower triagular indices
		tril_indices = th.tril_indices(row=self.action_size, col=self.action_size, offset=0)  

		# fill matrix with entries
		L[:, tril_indices[0], tril_indices[1]] = entries
		L.diagonal(dim1=1,dim2=2).exp_()

		# calculate state-dependent, positive-definite square matrix
		P = L*L.transpose(2, 1)

		"""
		Currently doesnt work with action spaces greater than size 1, need to fix!!
		"""
		
		Q = None
		if action is not None: 
			# assert action.shape == [N, 1 ,1]

			# action = action.transpose(2, 1)
			action_value = action_value.transpose(2, 1)
			print(action.shape, action_value.shape)

			# calculate Advantage:
			# A = (-0.5 * th.matmul(th.matmul((action - action_value).transpose(2, 1), P), (action - action_value))).squeeze(-1)
			A = (-0.5 * th.matmul(th.matmul((action - action_value).transpose(2, 1), P), (action - action_value))).squeeze(-1)

			print(A.shape, V.shape)	# problem is A
			Q = A + V   
		
		P = P.mean(0)   # Somewhere along the calculation P's size gets messes up, need to look into   

		# print(action_value.shape, entries.shape, V.shape)
		dist = th.distributions.MultivariateNormal(action_value.squeeze(-1), th.inverse(P))
		action = dist.sample()
		action = th.clamp(action, min=-1, max=1)
		return action, Q, V, action_value

net = NAF(state_size=8, action_size=2, layer_size=128)
targ_net = copy.deepcopy(net)
optimizer = th.optim.Adam(net.parameters(), lr=1e-3)

tau = 5e-4

for t in range(50000):
	batch = runner.get_batch(net)

	s, a, r, s_p, d = batch()

	_, q, _, _ = net(s, a)

	with th.no_grad():
		a_p, _, _, _ = targ_net(s_p)	# (1)
		
		# _, q_p, v_p, _ = targ_net(s_p, a_p.unsqueeze(-1))	# regular target q-net target computation

		_, q_p, v_p, _ = net(s_p, a_p.unsqueeze(-1))	# double dqn like target computation, works way better

		y = r + 0.99 * q_p * (1 - d)	# (1)
		# y = r + 0.99 * v_p * (1 - d)

		# (1) this was my own additional tweak, need to experiment with further but it appears to peform better!

	loss = F.mse_loss(q, y)

	optimizer.zero_grad()
	loss.backward()
	th.nn.utils.clip_grad_norm_(net.parameters(), 1.)
	optimizer.step()
	
	net.reset_noise()
	targ_net.reset_noise()

	for targ_params, params in zip(targ_net.parameters(), net.parameters()):
		targ_params.data.copy_(params.data * tau + targ_params.data * (1.0 - tau))


from .transitions import (BaseTransition,
                          TorchTransition,
                          JaxTransition)

REGISTRY = {}

REGISTRY["numpy"] = BaseTransition
REGISTRY["pytorch"] = TorchTransition
REGISTRY["jax"] = JaxTransition


from .basic import (BasePolicy, 
                    WhitenoiseDeterministic, 
                    EpsilonArgmax, 
                    Gaussian, 
                    NoisyNaf,
                    Categorical,
                    Beta)

REGISTRY = {}

REGISTRY["random"] = BasePolicy
REGISTRY["whitenoise"] = WhitenoiseDeterministic
REGISTRY["pinknoise"] = None
REGISTRY["ounoise"] = None
REGISTRY["argmax"] = EpsilonArgmax
REGISTRY["gaussian"] = Gaussian
REGISTRY["naf"] = NoisyNaf
REGISTRY["categorical"] = Categorical
REGISTRY["beta"] = Beta



import random
from cardio_rl.transitions import TorchTransition
from collections import deque


class ErTable:
	def __init__(self, capacity=1_000_000, transition_func=TorchTransition):
		self.buffer = deque(maxlen=capacity)
		self.transition_func = transition_func
	
	def __len__(self):
		return len(self.buffer)

	def store(self, transition):
		self.buffer.append(transition)

	def sample(self, batch_size):
		batch = random.sample(list(self.buffer), int(min(batch_size, self.__len__())))
		batch = self.transition_func(*zip(*batch))
		return batch()



import logging
from gymnasium import Env
from .transitions import REGISTRY as tran_REGISTRY
from .transitions import BaseTransition
from .buffers.er_buffer import ErTable
from .policies import BasePolicy, REGISTRY as pol_REGISTRY
from cardio_rl.gatherer import Collector

# https://stackoverflow.com/questions/40181284/how-to-get-random-sample-from-deque-in-python-3
# faster replay memory

class Runner():
    def __init__(
            self,
            env: Env,
            policy: BasePolicy = None,
            capacity: int = None,
            er_buffer: ErTable = None,
            batch_size: int = 100,
            collector: Collector = Collector(),
            n_batches: int = 1,
            reduce=True,
            backend = 'numpy',
        ) -> None:

        # Can maybe remove environment as an argument of the runner
        self.env = env

        # Maybe combine sampler and capacity into one argument?
        self.batch_size = batch_size
        self.n_batches = n_batches

        if er_buffer == None and capacity != None:
            transition = self._set_up_transition(backend)
            self.er_buffer = ErTable(capacity, transition)
            self.sampler = True
        elif capacity == None:            
            self.sampler = False
        else:
            self.er_buffer = er_buffer
            self.sampler = True

        self.collector = collector
        self.rollout_len = collector.rollout_len
        self.warmup_len = collector.warmup_len
        self.n_step = collector.n_step

        self.reduce = reduce
        self.backend = backend        

        self.policy = self._set_up_policy(policy)
        self.collector._init_policy(self.policy)
        self.collector._init_env(self.env)
        self._warm_start()            

    def _warm_start(
            self,
            net=None,          
        ):

        batch = self.collector.warmup(net)

        if self.sampler:
            for transition in batch:
                self.er_buffer.store(transition)
        
        logging.info('### Warm up finished ###')
        pass

    def _set_up_policy(self, policy):
        if isinstance(policy, str):
            return pol_REGISTRY[policy](self.env)

        elif isinstance(policy, BasePolicy):
            return policy

        else:
            # add warning
            return 
    
    def _set_up_transition(self, backend):
        """
        Maybe change name of backend to transition_type, better discription
        """
        if isinstance(backend, str):
            return tran_REGISTRY[backend]

        # isinstance(A, B) didn't work, this is a temp workaround
        elif backend.__base__ is BaseTransition:
            return backend

        else:
            # add warning
            return 

    def get_batch(
            self,
            net,
        ):
        
        self.net = net        
        rollout_batch = self.collector.rollout(net)

        if not self.sampler:            
            return self.prep_batch(rollout_batch)

        else:
            for transition in rollout_batch:
                self.er_buffer.store(transition)

            k = min(self.batch_size, len(self.er_buffer))
            
            if self.n_batches == 1:		
                return self.prep_batch(self.er_buffer.sample(k))

            else:
                batch_samples = []
                for _ in range(self.n_batches):
                    batch_samples.append(self.prep_batch(self.er_buffer.sample(k))) 

                return batch_samples


    def prep_batch(
        self,
        batch
        ):
        """
        takes the batch (which will be a list of transitons) and processes them to be seperate etc.
        """
        # need to redo after implementing replay buffer class

        if self.n_step == 1:
            return batch

        elif self.reduce == False:
            processed_batch = []
            for n_step_transition in batch:
                transition = n_step_transition
                processed_batch.append([*transition])

            return processed_batch
        
        else:
            processed_batch = []
            for n_step_transition in batch:
                s, a, r, s_p, d, i = n_step_transition
                s = s[0]
                a = a[0]
                r_list = list(r)
                s_p = s_p[-1]
                d = any(d)
                i = i
                processed_batch.append([s, a, r_list, s_p, d, i])

            return self.transition(*zip(*processed_batch))
        


from collections import deque
from gymnasium import Env
from cardio_rl.policies import BasePolicy
from cardio_rl.logger import Logger
import numpy as np
import gymnasium as gym

class Collector():
    def __init__(
            self,
            rollout_len: int = 1,
            warmup_len: int = 1_000,
            n_step: int = 1,
            take_every: int = 1,
            logger_kwargs = dict(
                    log_interval = 5_000,
                    episode_window=50,
                    tensorboard=False,
                )
        ) -> None:        

        self.rollout_len = rollout_len        

        if self.rollout_len == -1:
            self.ret_if_term = True
            self.rollout_len += 1000000
        else: 
            self.ret_if_term = False   

        self.warmup_len = warmup_len
        self.n_step = n_step   
        self.take_every = take_every 
        self.take_count = 0

        if logger_kwargs:
            self.logger = Logger(**logger_kwargs)
        else:
            self.logger = Logger()

        # env initialisation
        self.step_buffer = deque(maxlen=self.n_step)
    
    def _init_env(self, env: Env):
        self.env = env
        self.state, _ = self.env.reset()

    def _init_policy(self, policy):
        self.policy = policy

    def _env_step(self, policy=None, warmup=False):
        if warmup:
            a = self.env.action_space.sample()
        else:
            a = policy(self.state, self.net)
        s_p, r, d, t, info = self.env.step(a)
        self.logger.step(r, d, t)
        d = d or t
        return (self.state, a, r, s_p, d, info), s_p, d, t

    def warmup(            
        self,
        net = None,
    ):        
        # Maybe move this check to the runner?
        if self.warmup_len == None:
            return list(deque())
        
        self.net = net
        gather_buffer = deque()
        self.step_buffer = deque(maxlen=self.n_step)     

        for _ in range(self.warmup_len):
            transition, next_state, done, trun = self._env_step(self.policy, warmup=True)
            self.step_buffer.append(transition)
            
            if len(self.step_buffer) == self.n_step:
                
                if self.take_count % self.take_every == 0:                    
                    if self.n_step == 1:
                        gather_buffer.append(*list(self.step_buffer))
                    else:
                        gather_buffer.append(list(self.step_buffer))
 
                self.take_count += 1

            self.state = next_state
            if done or trun:
                self.state, _ = self.env.reset()
                self.step_buffer = deque(maxlen=self.n_step)

        return list(gather_buffer)

    def rollout(
        self,
        net,
    ):        
        self.net = net
        gather_buffer = deque()   

        for _ in range(self.rollout_len):
            transition, next_state, done, trun = self._env_step(self.policy)
            self.step_buffer.append(transition)
            if len(self.step_buffer) == self.n_step:
                
                if self.take_count % self.take_every == 0:
                    if self.n_step == 1:
                        gather_buffer.append(*list(self.step_buffer))
                    else:
                        gather_buffer.append(list(self.step_buffer))
 
                self.take_count += 1

            self.state = next_state
            if done or trun:
                self.state, _ = self.env.reset()
                self.step_buffer = deque(maxlen=self.n_step)
                self.policy.reset()
                if self.ret_if_term:
                    return list(gather_buffer)

        return list(gather_buffer)


class VectorCollector():
    def __init__(
            self,
            env,
            num_envs = 2,
            rollout_len = 1,
            warmup_len = 0,
            n_step = 1,
            logger_kwargs = None
        ) -> None:  
        
        # https://gymnasium.farama.org/api/vector/#async-vector-env            
        env_list = [lambda: gym.wrappers.RecordEpisodeStatistics(env) for _ in range(num_envs)]
        # Getting errors when using AsyncVectorEnv
        self.env = gym.vector.SyncVectorEnv(env_list)
        # self.env = gym.wrappers.RecordEpisodeStatistics(env)
        
        self.state, _ = self.env.reset()        

        self.rollout_len = rollout_len        

        if self.rollout_len == -1:
            self.ret_if_term = True
            self.rollout_len += 1000000
        else: 
            self.ret_if_term = False   

        self.warmup_len = warmup_len
        self.n_step = n_step        

        # metrics
        if logger_kwargs:
            self.logger = Logger(n_envs=num_envs, **logger_kwargs)
        else:
            self.logger = Logger(n_envs=num_envs)

    def init_policy(self, policy):
        self.policy = policy

    def _env_step(self, policy):
        a = policy(self.state, self.net)
        s_p, r, d, t, info = self.env.step(a)
        self.logger.vector_step(r, d, t)
        d_combined = [False] * len(d)

        for i in range(len(d_combined)):
            d_combined[i] = d[i] or t[i]
            
        return (self.state, a, r, s_p, d_combined, info), s_p, d, t

    def warmup(            
        self,
        net = None,
        policy = None,
    ):
        
        # Maybe move this check to the runner?
        if self.warmup_len == None:
            return deque()
        
        warmup_policy = BasePolicy(self.env)
        
        self.net = net
        gather_buffer = deque()
        step_buffer = deque(maxlen=self.n_step)     

        for _ in range(self.warmup_len):
            self.total_steps += 1
            a = policy(self.state, self.net)
            s_p, r, d, t, info = self.env.step(a)

            step_buffer.append([self.state, a, r, s_p, d])
            if len(step_buffer) == self.n_step:

                if self.n_step == 1:
                    gather_buffer.append(*list(step_buffer))
                else:
                    gather_buffer.append(list(step_buffer))

            self.state = s_p

        return list(gather_buffer)

    def rollout(
        self,
        net,
        policy,
    ):        
        self.net = net
        gather_buffer = deque()
        step_buffer = deque(maxlen=self.n_step)   

        for _ in range(self.rollout_len):
            transition, next_state, _, _ = self._env_step(policy)
            step_buffer.append(transition)
            if len(step_buffer) == self.n_step:
                if self.n_step == 1:
                    gather_buffer.append(*list(step_buffer))
                else:
                    gather_buffer.append(list(step_buffer))

            self.state = next_state

        return list(gather_buffer)


from typing import NamedTuple
import torch as th
import numpy as np
import jax.numpy as jnp

"""
All transition functions should output each vairable as shape [N, d]
where N is the batch size and d is the dimension of that specific variable,
i.e. d = [N, 1], r = [N, 1], s = [N, obs_dimensions]
"""

"""
Transitions need to be revisited some time later, I got extermely 
confused when trying to use them so it is likely unintuitive, each batch ends up
having variables named s,a,r,s_p,d but you need to call the batch function itself to
convert them to their respective back end? Maybe have the backend converstion as a method
of the Transition class itself? That could be a good fix
"""

class BaseTransition(NamedTuple):
    s: list # state
    a: float    # action
    r: float    # reward
    s_p: list   # next state
    d: int  # done
    i: dict #info
    
    def __call__(self):
        s = self.s
        a = self.a
        r = self.r
        s_p = self.s_p
        d = self.d
        i = self.i
        return s, a, r, s_p, d, i

class TorchTransition(BaseTransition):
    def __call__(self):
        s = th.from_numpy(np.array(self.s)).float()
        a = th.from_numpy(np.array(self.a)).unsqueeze(1).float()
        r = th.from_numpy(np.array(self.r)).unsqueeze(1).float()
        s_p = th.from_numpy(np.array(self.s_p)).float()
        d = th.from_numpy(np.array(self.d)).unsqueeze(1).int()
        return s, a, r, s_p, d, self.i
    
class JaxTransition(BaseTransition):
    def __call__(self):
        s = jnp.asarray(np.array(self.s))
        a = jnp.asarray(np.array(self.a))
        r = jnp.expand_dims(jnp.asarray(np.array(self.r)), -1)
        s_p = jnp.asarray(np.array(self.s_p))
        d = jnp.expand_dims(jnp.asarray(np.array(self.d)), -1)
        return s, a, r, s_p, d, self.i
    


from collections import deque
from torch.utils.tensorboard import SummaryWriter
import time
import logging
import numpy as np
from datetime import datetime

"""
-Maybe change to Rich Logger
-implement tensorboard next
-make logger compatible with VectorCollector next

# https://docs.python.org/3/howto/logging.html
# https://www.tensorflow.org/tensorboard
# https://pytorch.org/docs/stable/tensorboard.html
"""

class Logger():
    def __init__(
            self,
            n_envs = 1,
            log_interval=2000,
            episode_window=20,
            tensorboard=False,
            log_dir=None,
            exp_name='exp'
        ) -> None:
        
        logging.basicConfig(format='%(asctime)s: %(message)s', datefmt=' %I:%M:%S %p', level=logging.INFO)
        self.n_envs = n_envs
        self.log_interval = log_interval
        # self.episode_window = episode_window
        self.tensorboard = tensorboard

        if self.tensorboard:
            dir = ''

            if log_dir:
                dir += log_dir + '/'

            # time_key = str(int(time.time()//1))
            
            # Changed to the below to be more in line with Hydra
            date_key = datetime.now().strftime('%Y-%m-%d')
            time_key = datetime.now().strftime('%H-%M-%S')

            dir += exp_name + '_' + date_key + '_' + time_key + '/'
            self.writer = SummaryWriter(dir)
        
        self.timestep = 0
        self.episodes = 0
        if n_envs == 1:
            self.running_reward = 0
        else:
            self.running_reward = np.zeros(n_envs)

        self.episodic_rewards = deque(maxlen=episode_window)
        
    def step(self, reward, done, truncated):
        self.timestep += 1
        self.running_reward += reward

        if done or truncated:
            self.episodes += 1
            self.episodic_rewards.append(self.running_reward)
            self.running_reward = 0

        if self.timestep % self.log_interval == 0:
            logging.info(f'Timesteps: {self.timestep}, Episodes: {self.episodes}, Avg. reward is {np.mean(self.episodic_rewards)}')

            if self.tensorboard:
                self.writer.add_scalar('rollout/ep_rew_mean', np.mean(self.episodic_rewards), self.timestep)

    def vector_step(self, rewards, dones, truns):
        self.timestep += self.n_envs

        self.running_reward += rewards

        for i, (done, truncated) in enumerate(zip(dones, truns)):
            if done or truncated:
                self.episodes += 1
                self.episodic_rewards.append(self.running_reward[i])
                self.running_reward[i] = 0

        if (self.timestep//self.n_envs) % self.log_interval == 0:
            logging.info(f'Timesteps: {self.timestep}, Episodes: {self.episodes}, Avg. reward is {np.mean(self.episodic_rewards)}')




import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = 'cardio-rl',
    packages=find_packages(),
    install_requires=[
        "numpy>=1.21.0",
        "gymnasium>=0.28.0",
        'tensorboard'
    ],
    description = "Cardio RL. In development...",  
    author = 'Manus McAuliffe',
    url = 'https://github.com/mmcaulif/GymCardio',
    author_email = "mmcaulif@tcd.ie",  
    license = 'MIT',
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    version = '0.0.5',
    python_requires=">=3.7",
    
    # PyPI package information.
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



import torch as th
import numpy as np

class BasePolicy():
    def __init__(self, env, recurrent=False, hidden_dims=0):
        self.env = env        
        self.recurrent = recurrent
        self.hidden_dims = hidden_dims
        self.hidden = th.zeros(1, hidden_dims)
        
    def __call__(self, state, net):
        return self.env.action_space.sample()
    
    def reset(self):
        self.hidden = th.zeros(1, self.hidden_dims)
        # for policies that require resets, like NoisyNet or OuNoise
        # raise NotImplementedError
 
class WhitenoiseDeterministic(BasePolicy):
    def __init__(self, env):
        super().__init__(env)
        
    def __call__(self, state, net):
        input = th.from_numpy(state).float()  
        out = net(input)         
        mean = th.zeros_like(out)
        noise = th.normal(mean=mean, std=0.1)   # .clamp(-0.5, 0.5) # unsure if necessary... need to check other implementations
        out = (out + noise)    
        return out.clamp(-1, 1).detach().numpy()
                 

class EpsilonArgmax(BasePolicy):
    def __init__(self, env, eps=0.0, min_eps=0.0, ann_coeff=0.9, recurrent=False, hidden_dims=0):
        super().__init__(env, recurrent, hidden_dims)
        self.eps = eps
        self.min_eps = min_eps
        self.ann_coeff = ann_coeff

    def __call__(self, state, net):
        input = th.from_numpy(state).float()

        if np.random.rand() > self.eps:
            self.eps = max(self.min_eps, self.eps*self.ann_coeff)   
            if self.recurrent:
                out, self.hidden = net(input, hidden=self.hidden)
                out = out.detach().numpy()
            else: 
                out = net(input).detach().numpy()

            return np.argmax(out)  

        else:
            self.eps = max(self.min_eps, self.eps*self.ann_coeff)
            return self.env.action_space.sample()


class Gaussian(BasePolicy):
    def __init__(self, env):
        super().__init__(env)

    def __call__(self, state, net):
        input = th.from_numpy(state).float()
        mean, log_std = net(input)
        std = log_std.exp()

        dist = th.distributions.Normal(mean, std)
        a_sampled = th.nn.Tanh()(dist.rsample()).detach()

        return a_sampled.numpy() * self.env.action_space.high + 0


class NoisyNaf(BasePolicy):
    def __init__(self, env):
        super().__init__(env)

    def __call__(self, state, net):
        input = th.from_numpy(state).float()
        out, _, _, _ = net(input)
        return out.detach().numpy()


class Categorical(BasePolicy):
    def __init__(self, env):
        super().__init__(env)

    def __call__(self, state, net):
        input = th.from_numpy(state).float()
        probs = net(input)
        dist = th.distributions.Categorical(probs)
        return dist.sample().detach().numpy()
    
class Beta(BasePolicy):
    def __init__(self, env):
        super().__init__(env)

    def __call__(self, state, net):
        input = th.from_numpy(state).float()
        alpha, beta = net(input)
        dist = th.distributions.Beta(alpha, beta)
        return dist.sample().detach().numpy()